@@ -1,4 +1,5 @@
 import instancelib as il
+import pandas as pd
 from sklearn.feature_extraction.text import TfidfVectorizer
 from sklearn.naive_bayes import MultinomialNB
 
@@ -38,9 +39,9 @@ def test_build_from_model():
     train, test = env.train_test_split(env.dataset, 0.70)
     model = il.SkLearnVectorClassifier.build(MultinomialNB(), env)
     model.fit_provider(train, env.labels)
-    build_model = il.SkLearnVectorClassifier.build_from_model(model.innermodel, ints_as_str=True)
+    build_model = il.SkLearnVectorClassifier.build_from_model(model.innermodel, ints_as_str=True) # type: ignore
     model_preds = model.predict(test)
-    b_model_preds = build_model.predict(test)
+    b_model_preds = build_model.predict(test) # type: ignore
     for (a, pred_lt), (b, pred_idx) in zip(model_preds, b_model_preds):
         assert a == b
         first_label_lt = next(iter(pred_lt))
@@ -59,4 +60,37 @@ def test_confmat():
     conf_mat = il.analysis.confusion_matrix(env.labels, preds, test)
     print(conf_mat)
 
+def test_pandas_multiple():
+    df = pd.read_excel(DATASET_FILE)
+    env = il.pandas_to_env({"train": df, "test": df}, ["fulltext"], ["label"])
+    vect = il.TextInstanceVectorizer(il.SklearnVectorizer(TfidfVectorizer(max_features=1000)))
+    il.vectorize(vect, env)
+    assert env["train"]
+    assert env["test"]
+    assert env["train"]["train_20"].data == env["test"]["test_20"].data
+    assert env.labels.get_labels(env["train"]["train_20"])
+    assert env.labels.get_labels(env["train"]["train_20"]) == env.labels.get_labels(env["test"]["test_20"])
+    assert env.labels.get_labels(env["train"]["train_20"]) == frozenset({"Games"})
+    assert len(env["train"]) == len(env["test"])
+    train, test = env.train_test_split(env.dataset, 0.70)
+    model = il.SkLearnVectorClassifier.build(MultinomialNB(), env)
+    model.fit_provider(train, env.labels)
+    model_preds = model.predict(test)
+    preds = il.MemoryLabelProvider.from_tuples(model_preds)
+    conf_mat = il.analysis.confusion_matrix(env.labels, preds, test)
+    print(conf_mat)
+
+def test_pandas_single():
+    df = pd.read_excel(DATASET_FILE)
+    env = il.pandas_to_env(df, ["fulltext"], ["label"])
+    vect = il.TextInstanceVectorizer(il.SklearnVectorizer(TfidfVectorizer(max_features=1000)))
+    il.vectorize(vect, env)
+    train, test = env.train_test_split(env.dataset, 0.70)
+    model = il.SkLearnVectorClassifier.build(MultinomialNB(), env)
+    model.fit_provider(train, env.labels)
+    model_preds = model.predict(test)
+    preds = il.MemoryLabelProvider.from_tuples(model_preds)
+    conf_mat = il.analysis.confusion_matrix(env.labels, preds, test)
+    print(conf_mat)
+
     
\ No newline at end of file
